/*
	Copyright (C) 2022 Guénal Davalan. All right reserved.
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <inttypes.h>
#include <signal.h>

#include "Arduino.h"
#include "interrupt.h"
#include "cpu.h"


char _interruptEnabled=1;
int _pipeInterrupt[2];
extern pthread_t mainThread;

void interrupts() {
	_interruptEnabled=1;
	for(int i=0; i<NB_IO; ++i) {
		if(cpu.itMask[i]) { cpu.itMask[i]=0; sendInterrupt(i); }
	}
}

void noInterrupts() {
	_interruptEnabled=0;
}

void sendInterrupt(uint8_t it) {
	fprintf(stderr, "** send interrupt ms:%u\n", millis());
	//pthread_kill(mainThread, SIGRTMIN);
	
	if(write(_pipeInterrupt[1], &it, 1)!=1) {
		perror("write interrupt");
	}
	//pthread_kill(mainThread, SIGRTMIN);

}

void *interruptThread(void *arg) {
	uint8_t it;
	while(read(_pipeInterrupt[0], &it, 1)==1) {
		if(!_interruptEnabled) 
			cpu.itMask[it]=1;
		else {
			cpu.itMask[it]=0;
			if(cpu.itVector[it]!=NULL) {
				fprintf(stderr, "run interrupt ms:%u\n", millis());
				//dlock();
				(*cpu.itVector[it])();
				//dunlock();
				//fprintf(stderr, "run interrupt ok\n");
			}
			uint8_t value=(cpu.ios[it].getValue()!=0);
			if(cpu.itMode[it]==LOW && value==LOW) sendInterrupt(it);
			if(cpu.itMode[it]==HIGH && value==HIGH)  sendInterrupt(it);
		}
	}

	exit(1);
}

void initInterrupt() {
    pthread_t tid;

	if(pipe(_pipeInterrupt)==-1) {
		perror("pipe");
		exit(1);
	}

    if(pthread_create(&tid, NULL, interruptThread, NULL) != 0) {
        fprintf(stderr," pthread_create rtcThead error !\n");
        exit(1);
    }

    pthread_detach(tid);
}
