/*
	Copyright (C) 2022 Guénal Davalan. All right reserved.
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include <unistd.h>
#include <sys/time.h>
#include <signal.h>

#include "Arduino.h"
#undef min
#undef max

#include "posix.h"
#include "posixSerial.h"
#include "interrupt.h"
#include "tone.h"
#include "log.h"
#include "x11.h"


extern void setup(void);
extern void loop(void);

CPU cpu;

pthread_t mainThread;

double speedup=1.0;
uint32_t spiTime;
uint32_t usleepMin;

void intHandler(int sig) {
		fprintf(stderr, "intHandler...\n");
		exit(0);
}

/*
void rtminHandler(int sig) {
		fprintf(stderr, "** rtHandler...\n");
		usleep(1000000);
		//pthread_yield();
}
*/

//TFT_eSPI tfte;

void setupError() {
	fprintf(stderr, "setupError...\n");
	/*
	tfte.begin();
	tfte.setRotation(atoi(getenv("ORIENTATION")));
    tfte.fillScreen(TFT_RED);
    */
}

uint32_t calibrateUsleep() {
	uint64_t d, d1,d2, s=0;
	int n=100;
	for(int i=0; i<n; ++i) {
		d1=getTime();
		usleep(0);
		d2=getTime();
		d=d2-d1;
		//fprintf(stderr, "  d:%lu\n", d);
		s+=d;
	}
	
	return s*1000/n;
}

void init() {
	fprintf(stderr, "main init 1...\n");
	mainThread=pthread_self();
	
	usleepMin=calibrateUsleep();
	
	fprintf(stderr, "usleepMin:%u\n", usleepMin);
		
	/*
	if(signal(SIGRTMIN, rtminHandler)==SIG_ERR) {
			perror("signal");
	}
	*/
	
	if (setvbuf(stdout, NULL, _IONBF, 0)) {
		perror("setvbuf");
		exit(1);
	}
	
	speedup=getEnvDouble("SPEEDUP", 1.0);
	usbConnect=getEnvInt("USB_CONNECT", 1);
	glogFlag=getEnvInt("LOG", 0); display->logFlag=glogFlag;
	spiTime=getEnvInt("SPI_TIME", 400);
	
	fprintf(stderr, "spiTime:%u\n", spiTime);
	
	
	fprintf(stderr, "speedup:%f\n", speedup);
	
	fprintf(stderr, "main init 2...\n");
	initTone();
	initTime();	
	initInterrupt();
	initBoard();
	initSerial();	
	
	
	//while(!windowPresent) usleep(10000);
		
	if(getenv("ERROR")!=NULL) {
		setupError();
		while(1) sleep(100);
	}
	
	//usleep(100000);
	/*
	if(pthread_setschedprio(pthread_self(), 0)!=0) {
		fprintf(stderr, "ERROR : pthread_setschedprio\n");
	}
	
	struct sched_param p={19};
	if(pthread_setschedparam(pthread_self(), SCHED_IDLE, &p) != 0) {
		fprintf(stderr, "ERROR : pthread_setschedparam\n");
	}
	*/
	
	
}

int _main(int argc, char **argv)
{	

	init();
	fprintf(stderr, "Start setup...\n");

	setup();
	
	while(1) {
		loop();
		usleep(10/speedup);
	}

	return 0;
}


