/*
	Copyright (C) 2022 Guénal Davalan. All right reserved.
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <X11/Xlib.h>
#include <math.h>
#include <string.h>

#include "Arduino.h"
#include "posix.h"
#include "interrupt.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>

#include <cstddef> // for NULL
//#include <iostream>

#include "Arduino.h"
#include "posix.h"
#include "interrupt.h"

extern CPU cpu;

#define TwoWire int
/*
class TwoWire {


};
*/

TwoWire Wire1;


#ifdef __cplusplus
extern "C" {
#endif


	// TIME

	struct timeval arduinoStartTime;
	unsigned long long arduinoStartMicros, arduinoStartMillis;

	void delayMicroseconds(unsigned int value) {
        usleep(value/speedup);
    }

	void delay(uint32_t value) {
		//fprintf(stderr, "delay : %f\n", (value*1000)/speedup);
		usleep((value*1000)/speedup);
	}

	uint32_t micros() {
		struct timeval tv;

		if(!initTimeDone) return 0;
		
		gettimeofday(&tv, NULL);
		unsigned long long t=(tv.tv_sec*1000000+tv.tv_usec)*speedup;
		return (t-arduinoStartMicros); //*speedup;
	}

	uint32_t millis() {
		struct timeval tv;
		
		if(!initTimeDone) return 0;

		gettimeofday(&tv, NULL);
		unsigned long long t=(tv.tv_sec*1000000+tv.tv_usec)*speedup;
		return ((t-arduinoStartMicros)/1000);
	}

	void initTime() {
		srand(time(0));
		
		gettimeofday(&arduinoStartTime, NULL);
		arduinoStartMicros=(arduinoStartTime.tv_sec*1000000+arduinoStartTime.tv_usec)*speedup;
		arduinoStartMillis=(arduinoStartMicros/1000)*speedup;
		
		initTimeDone=1;
	}

	// for internal use
	
	int64_t getMicros(struct timeval *t) {
		return t->tv_sec*1000000L+t->tv_usec;
	}

	int64_t getTime() {
		struct timeval t;

		gettimeofday(&t, NULL);
		return getMicros(&t);
	}

	void setMicros(struct timeval *t, int64_t micros) {
		t->tv_sec = micros/1000000;
		t->tv_usec = micros%1000000;
	}

	void addTime(struct timeval *result, struct timeval *t1, struct timeval *t2) {
		setMicros(result, getMicros(t2)-getMicros(t1));
	}

	void subTime(struct timeval *result, struct timeval *t1, struct timeval *t2){
		setMicros(result, getMicros(t1)-getMicros(t2));
	}


	// io
	
	void sendIt(pin_size_t pin) {
		fprintf(stderr, "IT send : %d\n", pin);
	}


	uint16_t analogRead(pin_size_t pin) {
		if(pin >= NB_IO) { 
			fprintf(stderr, "[ERROR] digitlRead : pin=%u (max=%u) !\n", pin, NB_IO-1); 
			return 0; 
		}
		return cpu.ios[pin].getValue();
		
	}
    void analogWrite(pin_size_t pin, uint8_t value) {
		if(pin >= NB_IO) { 
			fprintf(stderr, "[ERROR] analogWrite : pin=%u (max=%u) !\n", pin, NB_IO-1); 
		}
		cpu.ios[pin].setValue(value);
	}

	void attachInterrupt(pin_size_t pin, void (*f)(void), uint8_t mode) {
		cpu.itMode[pin]=mode;
		cpu.itVector[pin]=f;
		
		uint8_t value=(cpu.ios[pin].getValue()!=0);
		
		if(cpu.itMode[pin]==LOW && value==LOW) sendInterrupt(pin);
		if(cpu.itMode[pin]==HIGH && value==HIGH)  sendInterrupt(pin);
	}
	
	void detachInterrupt(pin_size_t pin) {
		cpu.itVector[pin]=NULL;
	}
	
	void yield() {
		//pthread_yield();
	}
	
	void sleep_enable (void) {}
	void sleep_disable (void) {}
	void sleep_cpu (void) {}
	void sleep_mode (void) {}
	void sleep_bod_disable (void) {}


	int getEnvInt(char *name, int defaultValue) {
		if(getenv(name)) {
			return atoi(getenv(name));
		}
		else return defaultValue;
	}

	double getEnvDouble(char *name, double defaultValue) {
		if(getenv(name)) {
			return atof(getenv(name));
		}
		else return defaultValue;
	}


#ifdef __cplusplus
}
#endif


