/*
  PosixSerial.cpp - Hardware serial library for Wiring
  Copyright (c) 2006 Nicholas Zambetti.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  Modified 23 November 2006 by David A. Mellis
  Modified 28 September 2010 by Mark Sproul
  Modified 14 August 2012 by Alarus
  Modified 3 December 2013 by Matthijs Kooijman
  
  04/04/22 : Modified (POSIX adaptation) by Guénal Davalan
  * 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "Arduino.h"
#include <unistd.h>
#include <fcntl.h>

#include "posixSerial.h"

int usbConnect=1;
PosixSerial Serial;

// Public Methods //////////////////////////////////////////////////////////////

void PosixSerial::begin(unsigned long baud, uint8_t config)
{
	int flags = fcntl(0, F_GETFL, 0);// then, set the O_NONBLOCK flag
	if(fcntl(0, F_SETFL, flags | O_NONBLOCK)==-1) {
		perror("fcntl"); exit(1);
	}
}

void PosixSerial::end()
{
  int flags = fcntl(0, F_GETFL, 0);// then, set the O_NONBLOCK flag
	if(fcntl(0, F_SETFL, flags ^ O_NONBLOCK)==-1) {
		perror("fcntl"); exit(1);
	}
}

int PosixSerial::available(void)
{
  //fprintf(stderr,"int PosixSerial::available(void)\n");
  if(peek()==-1) return 0;
  return 1;
}

int PosixSerial::peek(void)
{
	//fprintf(stderr,"int PosixSerial::peek(void)\n");
	int c=read();
	if(c!=-1) ungetc(c, stdin);
	return c;
}

int PosixSerial::read(void)
{
	int c=getchar();
	return c;
}

int PosixSerial::availableForWrite(void)
{  
  return 1024;
}

void PosixSerial::flush()
{
	fflush(stdout);
  // If we have never written a byte, no need to flush. This special
  // case is needed since there is no way to force the TXC (transmit
  // complete) bit to 1 during initialization
}

size_t PosixSerial::write(uint8_t c)
{
	putchar(c); 
  
  return 1;
}

void initSerial() {
}

