/*
    The MIT License (MIT)
    Author: Hongtai Liu (lht856@foxmail.com)
    Copyright (C) 2020  Seeed Technology Co.,Ltd.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.	
    
    04/04/22 : Modified (POSIX adaptation) by Guénal Davalan 
   
*/

#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include "RTC_SAMD51.h"
#include "posix.h"


void *rtcThread(void *arg) {
	static int64_t startMicros, t, d, nbTicks=0, period=1000000;
	RTC_SAMD51 *pRtc=(RTC_SAMD51* )arg;

	startMicros=getTime();

	while(1) {
		++nbTicks;
		t=getTime();
		
		d=startMicros+(nbTicks)*period-t;
	
		pRtc->_dateTime=pRtc->_dateTime+TimeSpan(1);

		DateTime now=pRtc->_dateTime;
		
		if(pRtc->RTC_callBack != NULL) {
			for(int i=0; i<2; ++i) {
				uint8_t s=0,m,r=0;
				m=pRtc->_alarmMatch[i];
				if (m && pRtc->_alarmMatch[i]) {
					DateTime a=pRtc->_alarmDateTime[i];
					
					if( (r=m&1) && (a.second() == now.second())) s|=r;
					if( (r=m&2) && (a.minute() == now.minute())) s|=r;
					if( (r=m&4) && (a.hour() == now.hour())) s|=r;
					if( (r=m&8) && (a.day() == now.day())) s|=r;
					if( (r=m&16) && (a.month() == now.month())) s|=r;
					if( (r=m&32) && (a.year() == now.year())) s|=r;
				}
								
				if(m && s==m) pRtc->RTC_callBack(1<<i);

			}
		}

		if(d<10) usleep(10); else usleep(d);
	}

}


RTC_SAMD51::RTC_SAMD51() {
	RTC_callBack=NULL;
	
	if(pthread_create(&_tid, NULL, rtcThread, this) != 0) {
		fprintf(stderr," pthread_create rtcThead error !\n");
		exit(1);
	}

	pthread_detach(_tid);
	
}

RTC_SAMD51::~RTC_SAMD51() {
	RTC_callBack=NULL;
	
	if(_tid)
		pthread_kill(_tid, SIGTERM);
	_tid=0;
	
}

bool RTC_SAMD51::begin()
{

    return true;
}

void RTC_SAMD51::adjust(const DateTime &dt)
{
    _dateTime=dt ;

}

DateTime RTC_SAMD51::now()
{
    return _dateTime;
}

DateTime RTC_SAMD51::alarm(uint8_t id)
{
    if (id >= 2) return DateTime();

    return _alarmDateTime[id];
}

void RTC_SAMD51::setAlarm(uint8_t id, const DateTime &dt)
{
	
    if (id >= 2) return;
	_alarmDateTime[id]=dt;

}

void RTC_SAMD51::enableAlarm(uint8_t id, Alarm_Match match)
{
    if (id >= 2) return;

	_alarmMatch[id]=match;

}

void RTC_SAMD51::disableAlarm(uint8_t id)
{
    if (id >= 2) return;

	_alarmMatch[id]=0;
}

void RTC_SAMD51::attachInterrupt(rtcCallBack callback)
{
    RTC_callBack = callback;
}

void RTC_SAMD51::detachInterrupt()
{
    RTC_callBack = NULL;
}

void RTC_SAMD51::standbyMode()
{
}

/* Attach peripheral clock to 32k oscillator and RTC*/
void RTC_SAMD51::configureClock()
{
}

/*
 * Private Utility Functions
 */

/* Configure the 32768Hz Oscillator */
void RTC_SAMD51::config32kOSC()
{
}

/* Synchronise the CLOCK register for reading*/
inline void RTC_SAMD51::RTCreadRequest()
{
}

/* Wait for sync in write operations */
inline bool RTC_SAMD51::RTCisSyncing(uint32_t flag)
{
	_fakeFlag=flag;
    return 0;
}

void RTC_SAMD51::RTCdisable()
{
}

void RTC_SAMD51::RTCenable()
{
}

void RTC_SAMD51::RTCreset()
{
}

void RTC_SAMD51::RTCresetRemove()
{
}


