/*
    The MIT License (MIT)
    Author: Hongtai Liu (lht856@foxmail.com)
    Copyright (C) 2020  Seeed Technology Co.,Ltd.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
*/

#ifndef __RTC_SAMD51_H__
#define __RTC_SAMD51_H__

#include <pthread.h>
#include "DateTime.h"

typedef void (*rtcCallBack)(uint32_t flag); 

// Default date & time after reset
#define DEFAULT_YEAR 2000 // 2000..2063
#define DEFAULT_MONTH 1   // 1..12
#define DEFAULT_DAY 1     // 1..31
#define DEFAULT_HOUR 0    // 1..23
#define DEFAULT_MINUTE 0  // 0..59
#define DEFAULT_SECOND 0  // 0..59

class RTC_SAMD51
{
public:

	rtcCallBack RTC_callBack;

	DateTime _dateTime;
	DateTime _alarmDateTime[2];
	uint8_t _alarmMatch[2];
	uint32_t _fakeFlag;
	pthread_t _tid=0;


    enum Alarm_Match : uint8_t // Should we have this enum or just use the identifiers from /component/rtc.h ?
    {
        MATCH_OFF = 0,                  // Never
        MATCH_SS = 1,                    // Every Minute
        MATCH_MMSS = 3,                // Every Hour
        MATCH_HHMMSS = 7,            // Every Day
        MATCH_DHHMMSS = 15,         // Every Month
        MATCH_MMDDHHMMSS = 31,    // Every Year
        MATCH_YYMMDDHHMMSS = 63 // Once, on a specific date and a specific time
    };

	RTC_SAMD51();
	~RTC_SAMD51();
    boolean begin();
    void adjust(const DateTime &dt);
    DateTime now();
    void setAlarm(uint8_t id, const DateTime &dt);
    DateTime alarm(uint8_t id);
    void enableAlarm(uint8_t id, Alarm_Match match);
    void disableAlarm(uint8_t id);
    void attachInterrupt(rtcCallBack callback);
    void detachInterrupt();
    void standbyMode();

private:
    void config32kOSC(void);
    void configureClock(void);
    void RTCreadRequest();
    bool RTCisSyncing(uint32_t flag);
    void RTCdisable();
    void RTCenable();
    void RTCreset();
    void RTCresetRemove();
};




#endif
